__doc__ = """
We define a Field class from which all ZFormulator fields derive.
"""

# Common Zope utilities
from Globals import HTMLFile, HTML
from Globals import Persistent

import OFS.SimpleItem             # a Field is a SimpleItem
import Acquisition                # Various acquisition types.
import AccessControl.Role         # Zope Security management.

class ZFormulatorError(Exception):
    """Trivially simple error message; used if something goes wrong
    internally in ZFormulator.
    """
    pass

class ValidationError(Exception):
    """An error that is raised if this field fails validation.
    """
    pass

class SimpleStore:
    """Instances will be used to store objects dynamically.
    """
    pass

class Field(
    OFS.SimpleItem.Item,
    Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager
    ):
    """Base class for all Fields

    Each field defines a number of tabs:
    o a property tab where the field's properties can be edited
    o a messages tab where the field's error messages (to the end user)
      can be set
    o a test tab where we can test if the field behaves as we'd like
    o a view tab to view the HTML that will be produced if this field is
      rendered
    o a security tab
    """

    # so code can easily check if something's a field
    isZFormulatorField = 1
    
    # the various tabs
    manage_options = (
	{'label' : 'Edit',     'action' : 'manage_main'},
	{'label' : 'Messages', 'action' : 'manage_messagesForm'},
	{'label' : 'Test',     'action' : 'manage_testForm'},
	{'label' : 'Security', 'action' : 'manage_access'},
	)   

    # display edit screen as main management screen
    manage_main = HTMLFile('fieldEdit', globals())
    # display messages screen
    manage_messagesForm = HTMLFile('fieldMessages', globals())
    # display test screen
    manage_testForm = HTMLFile('fieldTest', globals())
    # display test feedback form
    manage_testFeedbackForm = HTMLFile('fieldTestFeedback', globals())
    # show this is validation went wrong
    manage_testValidationErrorForm =  HTMLFile('fieldTestValidationError', 
					       globals())
    manage_editValidationErrorForm = HTMLFile('fieldEditValidationError',
					      globals())
    # permissions
    __ac_permissions__ = (
	('View Management Screens',     
	 ('manage_tabs', 
	  'manage_main',
	  'manage_testForm', 
	  'manage_test',
	  'manage_testFeedbackForm',
	  'manage_testValidationErrorForm')
	 ),
	('Change permissions',          
	 ('manage_access',)),
	('Change ZFields',              
	 ('manage_edit',
	  'manage_editValidationErrorForm',
	  'manage_messagesForm',
	  'manage_messages')),
	('View',                        
	 ('index_html', 
	  'render', 
	  '__str__',)),
	('Access contents information', 
	 ('process', 
	  'fieldIds', 
	  'sqlType')),
	)

    def __init__(self, id, title):
	"""Initialize
	id - the id of this field (used as name in database and with
             field_ prepended on form)
        title - the title of the field (used to describe field for form)
        """
	# standard zope info
	self.id = id
	self.title = title
	# name of field as appearing on a form
	self.fieldname = "field_%s" % self.id
	
	# simple storage for properties and message texts
	self.messages = SimpleStore()
	self.props = SimpleStore()
	
	# set order property
	self.props.order = 0
	
	# holds message info
	self.messageList = []
	self.messageDict = {}
	
	# field will be configured later
	self.sheet = None
    
    def sqltype(self):
	"""Type in SQL table.
	"""
	return "<Unknown; subclass should override sqltype()>"

    def _setSheet(self, sheet):
	"""Configure this field with a property field sheet.
	The sheet will be checked to make sure it is not defining fields
        for properties that does not exist.

	sheet - the sheet to use with this field
	"""
	for id in sheet.fieldIds:
	    if not hasattr(self.props, id):
		raise ZFormulatorError, "Sheet defines field with id %s, "\
		      "but this is undefined." % id
	self.sheet = sheet
    
    def fieldIds(self):
	"""Get field ids for properties of field.
	"""
	return self.sheet.fieldIds

    def __str__(self):
	"""Render this field.
	"""
	return self.index_html()

    def index_html(self):
	"""Render this field.

        Uses the 'default' value for the field if possible.
        Renders this value using DTML, meaning you can use DTML statements
        in the default property of a field.
	"""
	if hasattr(self.props, 'default'):
            # default value provided in field

            # This would have been code that enables use of DTML
            # in the default property.
            # Unfortunately this opens a security hole; we somehow
            # need to pass authentication information here? 
            # default = HTML(self.props.default).__call__(self)
            default = self.props.default
	else:
            # no default value provided in field, so default is empty string
	    default = ""
	return self.render(default)

    def render(self, value):
	"""Render this field.

	value - the value of the contents of the field

        Call this if you want to supply the default value by hand explicitly.
        Does not use the 'default' value for the field.
	"""
	return "[Field base class cannot be rendered]"

    def manage_edit(self, title, REQUEST=None): 
	"""Handle result from edit action.

	title - the new title of the field
	REQUEST - REQUEST information to change field properties
	"""
	# set new title
	self.title = title

	# first process input, possibly raising exceptions on
	# validation failure
	for field in self.sheet.fieldList:
	    try:
		field.process(REQUEST)
	    except ValidationError, msg_id:
		# if something goes wrong during validation, 
		# show in special form
		REQUEST.set("field", field)
		REQUEST.set("message", msg_id)
		return self.manage_editValidationErrorForm(self, REQUEST)

	# store property values in props object
	props = self.props
	for id in self.sheet.fieldIds:
	    setattr(props, id, REQUEST["field_%s" % id])
	self.props = props

	# done with editing
	if REQUEST is not None:
            message = "Properties changed"
            return self.manage_main(self, REQUEST, manage_tabs_message=message)
   
    def manage_messages(self, REQUEST):
	"""Handle results from manage_messagesForm.
	Set new message descriptions.

	REQUEST - REQUEST information to change message texts
	"""
	messages = self.messages
	for message_id in self.messageList:
	    setattr(messages, message_id, REQUEST["msg_%s" % message_id])
	self.messages = messages

	# give 'edited' message on top of messages view
	if REQUEST is not None:
            message = "Messages changed"
            return self.manage_messagesForm(self, REQUEST,
                                            manage_tabs_message=message)
	
    def manage_test(self, REQUEST):
	"""Handle results from manage_testForm.
	
	REQUEST - REQUEST information containing the filled in field
                  (name is id prefixed with 'field_')
	"""
	# process input
	try:
	    self.process(REQUEST)
	except ValidationError, msg_id:
	    # if something goes wrong during validation, show in special form
	    REQUEST.set("message", msg_id)
	    return self.manage_testValidationErrorForm(self, REQUEST)
	
	return self.manage_testFeedbackForm(self, REQUEST)

    def _addMessage(self, message_id, message_desc):
	"""Add a message to this field.

	message_id - the id of this message (used in raised exceptions)
	message_desc - the description of the message
	"""
	messageList = self.messageList
	messageDict = self.messageDict
	messages = self.messages

	# add message to list of messages
	messageList.append(message_id)
	# add message desc to dictionary
	messageDict[message_id] = message_desc
	# add message value to dictionary
	setattr(messages, message_id, "")
	
	self.messageList = messageList
	self.messageDict = messageDict
	self.messages = messages	
    
    def _setMessage(self, message_id, message_text):
	messages = self.messages
	setattr(messages, message_id, message_text)
	self.messages = messages

    def process(self, REQUEST):
	"""Process field input.

	REQUEST - should contain an entry of 'field_' + id

	May change the values in REQUEST (type marshalling).
	May raise validation errors.
	"""
	pass


