from Globals import HTMLFile, ImageFile
from OFS.ObjectManager import ObjectManager
from AccessControl.Role import RoleManager
from OFS import SimpleItem
import Acquisition
import OFS
import string
import os
from Field import ValidationError

class Form(ObjectManager, RoleManager, SimpleItem.Item):
    """A form that holds formulator fields.
    """
    meta_type= "ZFormulatorForm"
    
    # define tabs
    manage_options= (
        {'label':'Contents', 'action':'manage_main'},
	{'label':'Test',     'action':'manage_testForm'},
	{'label':'SQL',      'action':'manage_sqlForm'},
	{'label':'Security', 'action':'manage_access'},
    )
   
    __ac_permissions__=(
        ('View management screens', 
	 ('manage_test',
	  'manage_testForm',
	  'manage_testFeedbackForm',
	  'manage_testValidationErrorForm',
	  'manage_sqlForm')),
        ('Access contents information',
         ('getFields',
	  'getFieldsInOrder',
	  'getSQLCreate',
	  'getSQLInsert'),
         ('Anonymous', 'Manager'),
         ),
	('Change ZFormulatorForms',
	 ('manage_addField',)),
    )

    # tab forms
    manage_testForm = HTMLFile('formTest', globals())
    manage_sqlForm = HTMLFile('formSQL', globals())
    
    # other forms
    manage_testFeedbackForm = HTMLFile('formTestFeedback', 
				       globals())
    manage_testValidationErrorForm = HTMLFile('formTestValidationError', 
					      globals())

    # meta_types will be filled by registerField function
    # thanks to the Other Martijn for the meta_types explanation!
    meta_types = ()

    def all_meta_types(self):
	"""We only want meta_types to be addable, not any acquired types.
	"""
	return self.meta_types
	
    def manage_addField(self, id, title, REQUEST):
	"""Add a new field; REQUEST has the fieldname of the field to add.
	self - the folder to which we want to add the field
	id - the id of the field to add
	title - the title of the field to add
	REQUEST - contains a 'fieldname' entry containing the name of the 
                  field that we want to add.
        """
	# create instance of the right field, using factory object
	field = factory.createInstance(REQUEST.fieldname, id, title)
	# add it to the form
	self._setObject(id, field)
	# return management screen
	return self.manage_main(self, REQUEST)

    def manage_test(self, REQUEST):
	"""Process test tab input for testing.
	"""
	try:
	    for field in self.getFields():
		field.process(REQUEST)
	except ValidationError, msg_id:
	    # if something goes wrong during validation, show in special form
	    REQUEST.set("message", msg_id)
	    REQUEST.set("field", field)
	    return self.manage_testValidationErrorForm(self, REQUEST)
	# show feedback form if all goes well
	return self.manage_testFeedbackForm(self, REQUEST)

    def validate(self, REQUEST):
	"""Validate form.
	"""
	error_messages = []
	
	for field in self.getFieldsInOrder():
	    try:
		field.process(REQUEST)
	    except ValidationError, msg_id:
		error_messages.append(getattr(field.messages, str(msg_id))) 
	REQUEST.set("formulator_errors", error_messages)
	
	    
    def getFields(self):
	"""Get all fields on this form (in no particular order).
	"""
	fields = []
	for obj in self.objectValues():
	    if obj.isZFormulatorField:
		fields.append(obj)
	return fields

    def getFieldsInOrder(self):
	"""Get all fields on this form, in order by 'order' property.
	"""
	fields = self.getFields()
	fields.sort(orderhelper)
	return fields

    def getSQLCreate(self):
	"""Get SQL create statement for this form.
	"""
	# get fields
	fields = self.getFieldsInOrder()

	# create list of types
	typeslist = []
	for field in fields:
	    typeslist.append("%s %s" % (field.id, field.sqltype()))
	typesstr = string.join(typeslist, ",\n")

	# construct create statement
	sql =\
"""
create table %s (
%s
);
""" % (self.id, typesstr)
	
	return sql
    
    def getSQLInsert(self):
	"""Get SQL insert statement for this form.
	"""
	fields = self.getFieldsInOrder()
	
	# create fieldnames and fieldvalues
	fieldnames = []
	fieldvalues = []
	for field in fields:
	    fieldnames.append(field.id)
	    fieldvalues.append('<dtml-sqlvar %s type="string">' %
			       field.fieldname)

	fieldnames_str = string.join(fieldnames, ",\n")
	fieldvalues_str = string.join(fieldvalues, ",\n")
	
	sql =\
"""
insert into %s (
%s
)
values (
%s
); 
""" % (self.id, fieldnames_str, fieldvalues_str)
	
	return sql

def orderhelper(a, b):
    return cmp(a.props.order, b.props.order)

# form to add the form to a folder
manage_addformForm = HTMLFile('formAdd', globals())

# adds form to folder
def manage_addform(self, id, title='', REQUEST=None):
    """Add form to folder.
    """
    form = Form()
    form.id = id
    form.title = title

    self._setObject(id, form)
    if REQUEST is not None:
	return self.manage_main(self, REQUEST)

class FieldFactory:
    """A helper class used to register fields with the Form.
    """
    def __init__(self):
	# a dictionary containing the field classes that have been added
	self.fieldclass_dict = {}

    def registerField(self, fieldclass, sheet):
	"""Register a fieldclass and associated sheet with the factory.
	
	fieldclass - a field class. The meta_type should be the name of
                     the field in Zope.
        sheet - a sheet instance that will be associated with each 
                instance of the field class that is created by Zope's
                machinery. Field instances created directly in
                Python won't have an associated sheet.
        """
	# add field info to dictionary
	self.fieldclass_dict[fieldclass.meta_type] = fieldclass, sheet

    def getFields(self):
	"""Get a list of registered fields by fieldname.
	"""
	return self.fieldclass_dict.keys()

    def createInstance(self, fieldname, id, title):
	"""Create an instance by fieldname.

	fieldname - name of the field we want to create an instance of
	id - the id of this instance
	title - the title of this instnace
	"""
	# get the right class and sheet
	fieldclass, sheet = self.fieldclass_dict[fieldname]
	# create instance
	instance = fieldclass(id, title)
	# set sheet in instance
	instance._setSheet(sheet)
	# return instance
	return instance

# create a global factory to register fields with
factory = FieldFactory()

def registerField(fieldclass, sheet, icon=None):
    """Register a field type with the ZFormulatorForm.
    fieldclass - the class of the field type to register
    sheet - the sheet of the field type to register
    icon  - icon of the class (None if no icon)

    NOTE: lots of Zope trickery to make this work
    """
    fieldname = fieldclass.meta_type

    # register with factory
    factory.registerField(fieldclass, sheet)

    # add this as an addable field to the form
    # for some reason we must use this tuple adding, 
    # as setting meta_types to an empty list in Form
    # does not appear work; meta_types always seems to be a tuple.
    Form.meta_types = Form.meta_types +\
	({ 'name': fieldname,
	  'action': 'manage_add%sForm' % fieldname },
	)

    # add manage_add method for this field to the Form class
    # use FormHack to create the right add form
    setattr(Form, 
            'manage_add%sForm' % fieldname,
    	     FormHack(fieldname))
    #setattr(Form,
    #        'manage_add%sForm' % fieldname,
    #	    HTMLFile('fieldAdd', globals()))
    
    # copy permissions to add special ones, lots of hackery ahead...
    permissions = Form.__ac_permissions__
    newpermissions = []
    for permission in permissions:
	# copy this permission to add new permission info
	if permission[0] == 'Change ZFormulatorForms':
	    change_permission = list(permission)
	else:
	    newpermissions.append(permission)
    # add the new information to the Change ZFormulatorForms permission
    methods = change_permission[1] + (('manage_add%sForm' % fieldname),)
    change_permission[1] = methods
    # add the new permission to the list
    newpermissions.append(tuple(change_permission))
    # now set the permissions in Form again
    Form.__ac_permissions__ = tuple(newpermissions)
    
    # trickery to set the icon 
    # hacked up code from lib/app/ProductContext.py
    if icon:
	# set up misc_ respository for ZFormulator if not existing yet
	if not hasattr(OFS.misc_.misc_, 'ZFormulator'):
	    setattr(OFS.misc_.misc_, 
		    'ZFormulator', 
		    OFS.misc_.Misc_('ZFormulator', {}))
	
	# get name of icon in the misc_ directory
	icon_name = os.path.split(icon)[1]
	
	# set up image object from icon file
	icon_image = ImageFile(icon, globals())
	icon_image.__roles__ = None

	# put icon image object in misc_/ZFormulator/
	getattr(OFS.misc_.misc_, 'ZFormulator')[icon_name] = icon_image

	# set icon attribute in fieldclass to point to this image obj
	setattr(fieldclass, 'icon', 'misc_/%s/%s' %
		('ZFormulator', icon_name))

class FormHack:
    """A real hack to make things work (HTMLFile doesn't work for some reason)
    """
    def __init__(self, fieldname):
	"""Create a form for fieldname.
	fieldname - the name of the field this form will create
	"""
	# configure the HTML of the form
	self.data = self.form_data % (fieldname, fieldname, fieldname)
    
    __roles__ = ('manager',)

    def __call__(self):
	"""Return the HTML contents of the form.
	"""
	return self.data

    # the HTML that should have been in a fieldAdd.dtml if HTMLFile had 
    # worked as I expected.
    form_data = '''
<html>
<head>
<title>Add %s</title>
</head>
<body bgcolor="#FFFFFF" link="#000099" vlink="#555555">

<h2>Add %s</h2>

<form action="manage_addField" method="POST">
  <table cellspacing="2">
    <tr>
      <th align="left" valign="top">Id</th>
      <td align="left" valign="top">
        <input type="text" name="id" size="40">
      </td>
    </tr>
    <tr>
      <th align="left" valign="top"><em>Title</em></th>
      <td align="left" valign="top">
        <input type="text" name="title" size="40">
      </td>
    </tr> 
    <tr>
      <td>
        <input type="hidden" name="fieldname" value="%s">
        <input type="submit" value=" OK ">
      </td>
    </tr>
  </table>
</form>

</body>
</html>
'''
    
